import pickle
import math
import os
from pathlib import Path
from .normalizer import Normalizer
from .tokenizer import Tokenizer
from .data_helper import DataHelper
import requests
import zipfile
from google_drive_downloader import GoogleDriveDownloader as gdd

class SpellCheck:
    def __init__(self):
        self.normalizer = Normalizer()
        self.tokenizer = Tokenizer()
        self.data_helper = DataHelper()

        self.dir_path = os.path.dirname(os.path.realpath(__file__)) + "/"

        my_file = Path(self.dir_path + "resource/spell")
        if my_file.is_file():

            self.bigram_lm = self.data_helper.load_var(self.dir_path + "resource/spell/created_bigram.pickle")
            self.onegram_lm = self.data_helper.load_var(self.dir_path + "resource/spell/created_onegram.pickle")
        else:
            gdd.download_file_from_google_drive(file_id='1cale2CmosneHP4y0MC_aW1f1g8dErjby',
                                                dest_path=self.dir_path + 'resource/spell.zip',
                                                unzip=True)

            self.bigram_lm = self.data_helper.load_var(self.dir_path + "resource/spell/created_bigram.pickle")
            self.onegram_lm = self.data_helper.load_var(self.dir_path + "resource/spell/created_onegram.pickle")

        self.ingroup_chars = [{'ا', 'آ', 'ع'},
                              {'ت', 'ط'},
                              {'ث', 'س', 'ص'},
                              {'ح', 'ه'},
                              {'ذ', 'ز', 'ض', 'ظ'},
                              {'ق', 'غ'}]

    def deletion(self, word):
        p_list = []
        for k in range(len(word)):
            if word[k] == '-' or word[k] == '#':
                continue
            begin = word[:k]
            end = word[k + 1:]
            tmp_string = begin + end
            p_list.append(tmp_string)
        return p_list

    def splitting(self, word):
        p_list = set([])
        delimator = '-'
        for i, char in enumerate(word):
            begin = word[:i].strip('\u200c')
            end = word[i:].strip('\u200c')
            tmp_string = begin + delimator + end
            p_list.add(tmp_string)
        return list(p_list)

    def insertion(self, word):
        p_list = []
        alphabet = ['ا', 'آ', 'ب', 'پ', 'ت', 'ث', 'ج', 'چ', 'ح', 'خ',
                    'د', 'ذ', 'ر', 'ز', 'ژ', 'س', 'ش', 'ص', 'ض', 'ط',
                    'ظ', 'ع', 'غ', 'ف', 'ق', 'ک', 'گ', 'ل', 'م', 'ن',
                    'و', 'ه', 'ی', '‌']
        for k in range(len(word) + 1):
            for char in alphabet:
                begin = word[:k]
                end = word[k:]
                tmp_string = begin + char + end
                p_list.append(tmp_string)
        return p_list

    def substitution(self, word):
        p_list = []
        alphabet = ['ا', 'آ', 'ب', 'پ', 'ت', 'ث', 'ج', 'چ', 'ح', 'خ',
                    'د', 'ذ', 'ر', 'ز', 'ژ', 'س', 'ش', 'ص', 'ض', 'ط',
                    'ظ', 'ع', 'غ', 'ف', 'ق', 'ک', 'گ', 'ل', 'م', 'ن',
                    'و', 'ه', 'ی']
        for i, char in enumerate(word):
            if char == '-' or char == '#':
                continue
            for c in alphabet:
                begin = word[:i]
                end = word[i + 1:]
                tmp_string = begin + c + end
                p_list.append(tmp_string)
        return p_list

    def transposition(self, word):
        p_list = []
        word = list(word)
        tmp_word = word[:]
        for k1 in range(len(word)):
            k2 = k1 + 1
            if k2 == len(word):
                break
            tmp = tmp_word[k1]
            tmp_word[k1] = tmp_word[k2]
            tmp_word[k2] = tmp
            tmp_string = ''.join(tmp_word)
            p_list.append(tmp_string)
            tmp_word = word[:]
        return p_list

    def build_similar_words(self, word_seq, index, zi, operation):
        z_list = []
        o_list = []

        if operation == "Spell":
            tmp = self.deletion(zi)
            for i in tmp:
                z_list.append(i)
                o_list.append("Deletion")
            tmp = self.insertion(zi)
            for i in tmp:
                z_list.append(i)
                o_list.append("Insertion")
            tmp = self.substitution(zi)
            for i in tmp:
                z_list.append(i)
                o_list.append("Substitution")
            tmp = self.transposition(zi)
            for i in tmp:
                z_list.append(i)
                o_list.append("Transposition")

        elif operation == "Split":
            tmp = self.splitting(zi)
            for i in tmp:
                z_list.append(i)
                o_list.append("Split")

        elif operation == "Merg":
            if index < len(word_seq) - 1:
                tmp = zi + '#' + word_seq[index + 1]
                z_list.append(tmp)
                o_list.append("Merg")

        return [z_list, o_list]

    def bigram_markov_factor(self, yi_1, yi):
        bigram_counts, total_count = self.bigram_lm
        tmp = (yi_1, yi)
        if tmp in bigram_counts.keys():
            x = bigram_counts[tmp]
            x = float(x) / total_count
            x = math.log2(x)
            return x
        else:
            return -28

    def get_word_probability(self, word):
        lex_dict = self.onegram_lm[0]
        total_words = self.onegram_lm[1]

        if word in lex_dict:
            count = lex_dict[word]
            logprob = math.log2(float(count) / total_words)
            return logprob
        else:
            return -50.0

    def isword(self, x):
        if abs(x.find('#') - x.find('-')) == 1:
            return False

        dash_idx = x.find('-')
        if dash_idx != -1:
            first = x[:dash_idx]  # from beginning to n (n not included)
            secound = x[dash_idx + 1:]  # n+1 through end of string
            if self.get_word_probability(first) < -49:
                return False
            elif self.get_word_probability(secound) < -49:
                return False
            else:
                return True

        sharp_idx = x.find('#')
        if sharp_idx != -1:
            begin = x[:sharp_idx]
            end = x[sharp_idx + 1:]
            tmp_str = begin + end
            if self.get_word_probability(tmp_str) < -49:
                return False
            else:
                return True
        else:
            if self.get_word_probability(x) < -49:
                return False
            else:
                return True

    def get_possible_words(self, word_seq, index):
        wi = word_seq[index]
        possible_words = []
        operation_list = []

        possible_words.append(wi)
        operation_list.append("Nothing")

        if len(wi) == 1:
            return possible_words, operation_list

        '''Merg Split Spell'''
        [c_list, o_list] = self.build_similar_words(word_seq, index, wi, "Merg")
        for i, c in enumerate(c_list):
            if self.isword(c):
                possible_words.append(c)
                operation_list.append(o_list[i])

        [c_list, o_list] = self.build_similar_words(word_seq, index, wi, "Split")
        for i, c in enumerate(c_list):
            if self.isword(c):
                possible_words.append(c)
                operation_list.append(o_list[i])

        # [c_list, o_list] = self.build_similar_words(word_seq, index, wi, "Spell")
        # for i, c in enumerate(c_list):
        #     if self.isword(c):
        #         possible_words.append(c)
        #         operation_list.append(o_list[i])

        return possible_words, operation_list

    def select_n_best(self, c_list, o_list, n=3):
        my_dict = {}
        map_dict = {}
        for i, word in enumerate(c_list):
            if o_list[i] == 'Merg':
                tmp_word = word.replace("#", "")
                prob = self.get_word_probability(tmp_word)

            elif o_list[i] == 'Split':
                begin = word.split('-')[0]
                end = word.split('-')[1]
                prob = float(self.get_word_probability(begin) + self.get_word_probability(end)) / 2

            else:
                prob = self.get_word_probability(word)

            if word not in my_dict:
                my_dict[word] = prob
                map_dict[word] = o_list[i]

        n_best = set(sorted(my_dict, key=my_dict.get, reverse=True)[:n])
        n_best.add(c_list[0])
        n_best = list(n_best)
        n_best_op = [map_dict[key] for key in n_best]
        return n_best, n_best_op

    def is_ingroup_substitution(self, main_word, candidate_word):
        main_word = list(main_word)
        candidate_word = list(candidate_word)
        flag = False
        for i, c in enumerate(main_word):
            if c == candidate_word[i]:
                continue
            else:
                flag = False
                for l in self.ingroup_chars:
                    if c in l and candidate_word[i] in l:
                        flag = True
                        break
                break
        return flag

    def select_correct_spell(self, candidate_list, next_candidates, next_next_candidates, prev_word, current_word):
        best_candidate = None
        best_operation = None
        best_score = -1000
        next_next_candidate_list = []
        next_next_operation_list = []

        candidate_list, operation_list = candidate_list
        if next_candidates is not None:
            next_candidate_list, next_operation_list = next_candidates
        else:
            next_candidate_list, next_operation_list = [None], "Nothing"

        if next_next_candidates is not None:
            next_next_candidate_list, next_next_operation_list = next_next_candidates
        else:
            next_candidate_list, next_operation_list = [None], "Nothing"

        for i, candidate in enumerate(candidate_list):
            operation = operation_list[i]

            if operation == "Split":
                begin = candidate[:candidate.find('-')]
                end = candidate[candidate.find('-') + 1:]
                candidate = begin
                next_word = end

                onegram_score = self.get_word_probability(candidate)
                bigram_score_with_prev = self.bigram_markov_factor(prev_word, candidate)

                bigram_score_next = -1000
                tmp_score_next = self.bigram_markov_factor(candidate, next_word)
                for j, next_next_word in enumerate(next_candidate_list):
                    opt = next_operation_list[j]
                    if opt == 'Merg':
                        next_next_word = next_next_word.replace("#", "")
                    elif opt == 'Split':
                        next_next_word = next_next_word.split('-')[0]

                    tmp_score_next_next = self.bigram_markov_factor(next_word, next_next_word)
                    if tmp_score_next_next > bigram_score_next:
                        bigram_score_next = tmp_score_next_next
                bigram_score_next = float(bigram_score_next + tmp_score_next) / 2

            elif operation == "Merg":
                begin = candidate[:candidate.find('#')]
                end = candidate[candidate.find('#') + 1:]
                candidate = begin + end

                onegram_score = self.get_word_probability(candidate)
                bigram_score_with_prev = self.bigram_markov_factor(prev_word, candidate)

                bigram_score_next = -1000
                for j, next_next_word in enumerate(next_next_candidate_list):
                    opt = next_next_operation_list[j]
                    if opt == 'Merg':
                        next_next_word = next_next_word.replace("#", "")
                    elif opt == 'Split':
                        next_next_word = next_next_word.split('-')[0]

                    tmp_score = self.bigram_markov_factor(candidate, next_next_word)
                    if tmp_score > bigram_score_next:
                        bigram_score_next = tmp_score

            else:
                onegram_score = self.get_word_probability(candidate)
                bigram_score_with_prev = self.bigram_markov_factor(prev_word, candidate)

                bigram_score_next = -1000
                for j, next_word in enumerate(next_candidate_list):
                    opt = next_operation_list[j]
                    if opt == 'Merg':
                        next_word = next_word.replace("#", "")
                    elif opt == 'Split':
                        next_word = next_word.split('-')[0]

                    tmp_score = self.bigram_markov_factor(candidate, next_word)
                    if tmp_score > bigram_score_next:
                        bigram_score_next = tmp_score

            if operation == 'Substitution':
                if self.is_ingroup_substitution(current_word, candidate):
                    onegram_score += 20
                else:
                    onegram_score += 10
            elif operation == 'Deletion' or operation == 'Insertion':
                onegram_score += 5
                if '\u200c' in candidate and '\u200c' not in current_word:
                    onegram_score += 5
            elif operation == 'Split' or operation == 'Merg':
                onegram_score += 7
            elif operation == 'Nothing':
                onegram_score += 20

            score = 1 * onegram_score + 0.7 * bigram_score_with_prev + 0.7 * bigram_score_next

            if score > best_score:
                best_operation = operation
                best_candidate = candidate_list[i]
                best_score = score

        return best_candidate, best_operation

    def spell_corrector(self, doc_string):
        words = self.tokenizer.tokenize_words(doc_string)

        best_o_list = []
        best_candidates_list = []

        yi_1 = None
        merged_before = False

        suggest_list = []

        for i, word in enumerate(words):
            [c_list, o_list] = self.get_possible_words(words, i)
            n_best = self.select_n_best(c_list, o_list, n=15)
            suggest_list.append(n_best)

        for i, candidate_list in enumerate(suggest_list):

            if merged_before:
                continue

            if (i + 2) < len(suggest_list):
                next_candidates = suggest_list[i + 1]
                next_next_candidates = suggest_list[i + 2]
            elif (i + 1) < len(suggest_list):
                next_candidates = suggest_list[i + 1]
                next_next_candidates = None
            else:
                next_candidates = None
                next_next_candidates = None

            best_candidate, best_operation = self.select_correct_spell(candidate_list, next_candidates,
                                                                       next_next_candidates, yi_1, words[i])

            merged_before = False
            if best_operation == "Split":
                begin = best_candidate.split('-')[0]
                end = best_candidate.split('-')[1]
                best_candidate = [begin, end]
            if best_operation == "Merg":
                best_candidate = best_operation.replace("#", "")
                merged_before = True
            if type(best_candidate) == str:
                best_candidate = [best_candidate]

            best_o_list.append(best_operation)
            best_candidates_list.extend(best_candidate)
            yi_1 = best_candidate[-1]

        res = " ".join(best_candidates_list)
        ops = " ".join(best_o_list)

        return res


if __name__ == "__main__":
    doc_string = "نمازگذاران وارد مسلی شدند."
    myspell_checker = SpellCheck()

    res = myspell_checker.spell_corrector(doc_string)
    print(res)



from spacy import Language

from .char_normalizer import Char_Normalizer
from .date2string import Date2String
from .time2string import Time2String
from .num2string import Num2String
from .abbreviation import Abbreviation
from .punctuation_remover import Punc_remover
from .affix_norm import AffixNorm
from .spell_checker import SpellCheck
from .tokenizer import Tokenizer as parsivar_tokenizer
from .mapping_words import Word_Mapper
import spacy
import os


# word and sentence tokenizer
# @Language.factory("disabletokenizer")
class WhitespaceTokenizer():
    def __init__(self, vocab):
        self.vacab = vocab

    def __call__(self, string):
        return string


# @spacy.registry.tokenizers("whitespace_tokenizer")
# def create_whitespace_tokenizer():
#     def create_tokenizer(nlp):
#         return WhitespaceTokenizer(nlp)
#
#     return create_tokenizer


# @Language.factory("CustomSentenceTokenizer")
# class CustomSentenceTokenizer():
#     def __init__(self):
#         self.tokenizer = parsivar_tokenizer()
#
#     def __call__(self, string):
#         return self.tokenizer.tokenize_sentences(string)
# nlp.tokenizer = disabletokenizer()
# nlp.tokenizer_sentence = CustomSentenceTokenizer()


# -------------------------------------------

class _normalizer:
    def __init__(self, nlp: Language, config_file: str):
        self.config_file = config_file
        self.normalizer = Char_Normalizer(config_file)

    def __call__(self, doc):
        return self.normalizer.normalize(doc)


@Language.factory("char_normalizer", default_config={"config_file": str})
def char_normalizer(nlp: Language, name: str, config_file: str):
    return _normalizer(nlp, config_file)


# -----------------------------------------------------------


class _date_2_string:
    def __init__(self, nlp: Language):
        self.date_2_string = Date2String()

    def __call__(self, doc):
        return self.date_2_string.normalize(doc)


@Language.factory("date2str", default_config={})
def date2str(nlp: Language, name: str):
    return _date_2_string(nlp)


# -----------------------------------------------------------
class _num_2_string:
    def __init__(self, nlp: Language):
        self.num_2_string = Num2String()

    def __call__(self, doc):
        return self.num_2_string.normalize(doc)


@Language.factory("num2str", default_config={})
def num2str(nlp: Language, name: str):
    return _num_2_string(nlp)


# -----------------------------------------------------------
class _time_2_string:
    def __init__(self, nlp: Language):
        self.time_2_string = Time2String()

    def __call__(self, doc):
        return self.time_2_string.normalize(doc)


@Language.factory("time2str", default_config={})
def time2str(nlp: Language, name: str):
    return _time_2_string(nlp)


# -----------------------------------------------------------
class _Abbreviation_2_word:
    def __init__(self, nlp: Language):
        self.Abbreviation_2_word = Abbreviation()

    def __call__(self, doc):
        return self.Abbreviation_2_word.normalize(doc)


@Language.factory("abbreviation2word", default_config={})
def Abbreviation2word(nlp: Language, name: str):
    return _Abbreviation_2_word(nlp)


# -----------------------------------------------------------
parsivar_tokenizer = parsivar_tokenizer()


@Language.component("word_level_tokenizer")
def word_level_tokenizer(doc):
    # print("tokenize2str Process Done!")
    return parsivar_tokenizer.tokenize_words(doc)


# -----------------------------------------------------------
class _punctuation_remover:
    def __init__(self, nlp: Language):
        self.punc_remover = Punc_remover()

    def __call__(self, doc):
        return self.punc_remover.normalize(doc)


@Language.factory("punctuation_remover", default_config={})
def punctuation_remover(nlp: Language, name: str):
    return _punctuation_remover(nlp)


# -----------------------------------------------------------

class _affix2norm:
    def __init__(self, nlp: Language, config_file: str):
        self.config_file = config_file
        self.affix = AffixNorm(self.config_file)

    def __call__(self, doc):
        return self.affix.normalize(doc)


@Language.factory("affix2norm", default_config={"config_file": str})
def affix2norm(nlp: Language, name: str, config_file: str):
    return _affix2norm(nlp, config_file)


# -----------------------------------------------------------

class _wordmapper:
    def __init__(self, nlp: Language, config_file: str, half_space_char: str):
        self.config_file = config_file
        self.half_space_char = half_space_char
        self.mapper = Word_Mapper(self.config_file, half_space_char)

    def __call__(self, doc):
        return self.mapper.normalize(doc)


@Language.factory("word_mapper", default_config={"config_file": str, "half_space_char": str})
def word_mapper(nlp: Language, name: str, config_file: str, half_space_char):
    return _wordmapper(nlp, config_file, half_space_char)


# -----------------------------------------------------------
class _spell_cheker:
    def __init__(self, nlp: Language):
        self.spell = SpellCheck()

    def __call__(self, doc):
        return self.spell.spell_corrector(doc)


@Language.factory("spell_checker", default_config={})
def spell_checker(nlp: Language, name: str):
    return _spell_cheker(nlp)


# -----------------------------------------------------------
nlp = spacy.blank("fa")

conf_path = os.path.dirname(os.path.realpath(__file__)) + "/config/"

nlp.tokenizer = WhitespaceTokenizer(nlp)
nlp.add_pipe("char_normalizer", first=True,
             config={"config_file": conf_path + 'character_mappings.json'})
nlp.add_pipe("spell_checker", after="char_normalizer", config={})
nlp.add_pipe("date2str", after="spell_checker", config={})
nlp.add_pipe("time2str", after="date2str", config={})
nlp.add_pipe("num2str", after="time2str", config={})
nlp.add_pipe("abbreviation2word", after="num2str", config={})
nlp.add_pipe("affix2norm", after="abbreviation2word",
             config={"config_file": conf_path + 'affix.json'})
nlp.add_pipe("word_mapper", after="affix2norm",
             config={"config_file": conf_path + 'word_mappings.json', "half_space_char": "  "})

nlp.add_pipe("punctuation_remover", after="abbreviation2word", config={})
nlp.add_pipe("word_level_tokenizer")

nlp_blank = spacy.blank("fa")
nlp_blank.tokenizer = WhitespaceTokenizer(nlp_blank)



from nltk.parse.dependencygraph import DependencyGraph
from nltk.parse.malt import MaltParser
import os
import tempfile
from .stemmer import FindStems
from .postagger import POSTagger
from .tokenizer import Tokenizer
from .normalizer import Normalizer


class MyMaltParser(MaltParser):
    def __init__(self, parser_dirname, model_filename, tagger, stemmer):
        """
        An interface for parsing with the Malt Parser.
        :param parser_dirname: The path to the maltparser directory that
        contains the maltparser-1.x.jar
        :type parser_dirname: str
        :param model_filename: The name of the pre-trained model with .mco file
        extension. If provided, training will not be required.
        (see http://www.maltparser.org/mco/mco.html and
        see http://www.patful.com/chalk/node/185)
        :type model_filename: str
        :param tagger: The tagger used to POS tag the raw string before
        formatting to CONLL format. It should behave like `nltk.pos_tag`
        :type tagger: function
        :param stemmer: a function which returns stem of the word
        :type function
        """
        self.working_dir = parser_dirname
        self.mco = model_filename
        self.pos_tagger = tagger
        self._malt_bin = os.path.join(parser_dirname, 'maltparser-1.9.2.jar')
        self.stemmer = stemmer.convert_to_stem if stemmer else lambda w, t: '_'

    def parse_tagged_sent(self, sentences, verbose=False, top_relation_label='null'):
        tmp_file_address = tempfile.gettempdir()
        input_file = tempfile.NamedTemporaryFile(prefix='malt_input.conll', dir=tmp_file_address, delete=False)
        output_file = tempfile.NamedTemporaryFile(prefix='malt_output.conll', dir=tmp_file_address, delete=False)

        for sentence in sentences:
            for i, (word, tag) in enumerate(sentence, start=1):
                word = word.strip()
                if not word:
                    word = '_'
                input_file.write(('\t'.join([str(i), word.replace(' ', '_'), self.stemmer(word, tag).replace(' ', '_'), tag, tag, '_', '0', 'ROOT', '_', '_', '\n'])).encode('utf8'))
            input_file.write('\n'.encode('utf8'))
        input_file.close()

        cmd = ['java', '-jar', self._malt_bin, '-w', self.working_dir, '-c', self.mco, '-i', input_file.name, '-o', output_file.name, '-m', 'parse']
        if self._execute(cmd, verbose) != 0:
            raise Exception("MaltParser parsing failed: %s" % (' '.join(cmd)))

        dependency_graph = []
        with open(output_file.name, encoding='utf-8') as infile:
            content = infile.read().strip().split('\n\n')
            for sent in content:
                dependency_graph.append(DependencyGraph(sent))

        input_file.close()
        output_file.close()
        os.remove(input_file.name)
        os.remove(output_file.name)
        return dependency_graph


class DependencyParser:
    def __init__(self, _normalizer=None, _tokenizer=None, _stemmer=None, _tagger=None):
        self.dir_path = os.path.dirname(os.path.realpath(__file__)) + "/"

        if _normalizer is None:
            self.my_normalizer = Normalizer()
        else:
            self.my_normalizer = _normalizer

        if _tokenizer is None:
            self.my_tokenizer = Tokenizer()
        else:
            self.my_tokenizer = _tokenizer

        if _stemmer is None:
            self.my_stemmer = FindStems()
        else:
            self.my_stemmer = _stemmer

        if _tagger is None:
            self.my_tagger = POSTagger(tagging_model="wapiti").parse
        else:
            self.my_tagger = _tagger

        self.parser = MyMaltParser(parser_dirname=self.dir_path + 'resource/dependency_parser',
                                   model_filename='total_dep_parser.mco',
                                   tagger=self.my_tagger,
                                   stemmer=self.my_stemmer)

    def make_trainable_corpus(self, in_file, out_file):
        tagger = self.my_tagger
        with open(in_file, 'r') as infile:
            content = infile.read().strip().split('\n\n')
            for i, sent in enumerate(content):
                if len(sent) == 0:
                    continue
                lines = sent.split('\n')
                sent_tokens = [x.split('\t')[1] for x in lines]
                tagged_sent = tagger(sent_tokens)
                tages = [x[1] for x in tagged_sent]
                for j, line in enumerate(lines):
                    line = line.split('\t')
                    line[3] = tages[j]
                    line[4] = tages[j]
                    line = '\t'.join(line)
                    lines[j] = line
                sent = '\n'.join(lines)
                content[i] = sent
        content = '\n\n'.join(content)
        with open(out_file, 'w') as outfile:
            outfile.write(content)
        return content

    def parse_sents(self, sents, verbose=False):
        tagger = self.my_tagger
        tagged_sents = [tagger(self.my_tokenizer.tokenize_words(sent)) for sent in sents]
        return self.parser.parse_tagged_sent(tagged_sents, verbose)



from re import sub
import string
import re
import json
import os
class Char_Normalizer():

    def __init__(self,config_file):

        self.punctuations = r')(}{:؟!،؛»«.' + r"/<>?.,:;"
        self.punctuations = '[' + self.punctuations + string.punctuation + ']'
        self.punctuations = self.punctuations.replace("@", "")
        self.pattern = '\s*' + self.punctuations + '+' + '\s*'

        self.dir_path = os.path.dirname(os.path.realpath(__file__)) + "/"



        with open(config_file) as f:
            mapper_tmp = json.load(f)
        self.char_mapper = {}
        for i in mapper_tmp['Mapper']:
            self.char_mapper.update(mapper_tmp['Mapper'][i])

        self.rule = "|".join(self.char_mapper.keys())
        self.rule = r"(" + self.rule + r")"

    def sub_alphabets(self, doc_string):

        text = re.sub(self.rule, lambda m: self.char_mapper.get(str(m.group()), str(m.group())), str(doc_string))
        return text

    def normalize(self, doc_string):

        doc_string = re.sub(self.pattern, self.add_space, doc_string)

        normalized_string = self.sub_alphabets(doc_string)

        return normalized_string

    def add_space(self, mystring):
        mystring = mystring.group()  # this method return the string matched by re
        mystring = mystring.strip(' ')  # ommiting the whitespace around the pucntuation
        mystring = " " + mystring + " "  # adding a space after and before punctuation
        return mystring



import os
import json
import re


class Word_Mapper:

    def __init__(self, config_file=os.path.dirname(os.path.realpath(__file__)) + "/config/word_mappings.json",
                 half_space_char='\u200c'
                 ):
        with open(config_file) as f:
            self.dictionary_mapping = json.load(f)

        self.half_space_char = half_space_char
        self.rule = "|".join(self.dictionary_mapping.keys())
        self.rule = r"(?<!\w)(" + self.rule + r")(?!\w)"

    def normalize(self, doc_string):
        text = re.sub(self.rule, lambda m: self.dictionary_mapping.get(str(m.group()), str(m.group())), str(doc_string))

        text = re.sub(r"\u200c", self.half_space_char, text)
        text = re.sub(r' +', ' ', text)

        return "".join(text.rstrip().lstrip())



from re import sub
import copy
import os
from .tokenizer import Tokenizer
from .data_helper import DataHelper
from .token_merger import ClassifierChunkParser
from num2fawords import words, ordinal_words
import re
from datetime import datetime
import random
import mofid_normalizer.stemmer
import nltk.corpus
import string

from num2fawords import words as internal_words


class Normalizer():

    def __init__(self,
                 half_space_char='\u200c',
                 statistical_space_correction=False,
                 date_normalizing_needed=False,
                 time_normalizing_nedded=False,
                 number_normalizing_needed=False,
                 pinglish_conversion_needed=False,
                 punctuation_normalizing_needed=False,
                 abbreviation_needed=False,
                 upper_case=False,
                 space_normalizing_needed=True,
                 train_file_path="resource/tokenizer/Bijan_khan_chunk.txt",
                 token_merger_path="resource/tokenizer/TokenMerger.pckl"):
        self.dir_path = os.path.dirname(os.path.realpath(__file__)) + "/"
        self.dic1_path =self.dir_path +'resource/normalizer/Dic1_new.txt'
        self.dic2_path =self.dir_path +'resource/normalizer/Dic2_new.txt'
        self.dic3_path =self.dir_path +'resource/normalizer/Dic3_new.txt'
        self.dic1 = self.load_dictionary(self.dic1_path)
        self.dic2 = self.load_dictionary(self.dic2_path)
        self.dic3 = self.load_dictionary(self.dic3_path)

        self.upper_case = upper_case
        self.statistical_space_correction = statistical_space_correction
        self.date_normalizing_needed = date_normalizing_needed
        self.punctuation_normalizing_needed = punctuation_normalizing_needed
        self.time_normalizing_needed = time_normalizing_nedded
        self.number_normalizing_needed = number_normalizing_needed
        self.space_normalizing_needed = space_normalizing_needed
        self.punctuation_normalizing_needed = punctuation_normalizing_needed
        self.abbreviation_needed = abbreviation_needed
        self.pinglish_conversion_needed = pinglish_conversion_needed
        self.data_helper = DataHelper()
        self.token_merger = ClassifierChunkParser()

        if self.date_normalizing_needed or self.pinglish_conversion_needed or self.time_normalizing_needed or self.number_normalizing_needed or self.abbreviation_needed or self.punctuation_normalizing_needed or self.space_normalizing_needed:
            self.tokenizer = Tokenizer()
            self.date_normalizer = DateNormalizer()
            self.time_normalizer = TimeNormalizer()
            self.pinglish_conversion = PinglishNormalizer()
            self.number_normalizer = NumberNormalizer()
            self.abbreviation_normalizer = AbbreviationNormalizer()
            self.punctuation_normalizer = Punctuation()
            self.space_normalizer = Spacenormalizer()

        if self.statistical_space_correction:
            self.token_merger_path = self.dir_path + token_merger_path
            self.train_file_path = train_file_path
            self.half_space_char = half_space_char

            if os.path.isfile(self.token_merger_path):
                self.token_merger_model = self.data_helper.load_var(self.token_merger_path)
            elif os.path.isfile(self.train_file_path):
                self.token_merger_model = self.token_merger.train_merger(self.train_file_path, test_split=0)
                self.data_helper.save_var(self.token_merger_path, self.token_merger_model)

    def load_dictionary(self, file_path):
        dict = {}
        with open(file_path, 'r', encoding='utf-8') as f:
            g = f.readlines()
            for Wrds in g:
                wrd = Wrds.split(' ')
                dict[wrd[0].strip()] = sub('\n', '', wrd[1].strip())
        return dict

    def sub_alphabets(self, doc_string):

        # try:
        #     doc_string = doc_string.decode('utf-8')
        # except UnicodeEncodeError:
        #     pass
        a0 = "ء"
        b0 = "ئ"
        c0 = sub(a0, b0, doc_string)
        a1 = r"ٲ|ٱ|إ|ﺍ|أ"
        a11 = r"ﺁ|آ"
        b1 = r"ا"
        b11 = r"آ"
        c11 = sub(a11, b11, c0)
        c1 = sub(a1, b1, c11)
        a2 = r"ﺐ|ﺏ|ﺑ"
        b2 = r"ب"
        c2 = sub(a2, b2, c1)
        a3 = r"ﭖ|ﭗ|ﭙ|ﺒ|ﭘ"
        b3 = r"پ"
        c3 = sub(a3, b3, c2)
        a4 = r"ﭡ|ٺ|ٹ|ﭞ|ٿ|ټ|ﺕ|ﺗ|ﺖ|ﺘ"
        b4 = r"ت"
        c4 = sub(a4, b4, c3)
        a5 = r"ﺙ|ﺛ"
        b5 = r"ث"
        c5 = sub(a5, b5, c4)
        a6 = r"ﺝ|ڃ|ﺠ|ﺟ"
        b6 = r"ج"
        c6 = sub(a6, b6, c5)
        a7 = r"ڃ|ﭽ|ﭼ"
        b7 = r"چ"
        c7 = sub(a7, b7, c6)
        a8 = r"ﺢ|ﺤ|څ|ځ|ﺣ"
        b8 = r"ح"
        c8 = sub(a8, b8, c7)
        a9 = r"ﺥ|ﺦ|ﺨ|ﺧ"
        b9 = r"خ"
        c9 = sub(a9, b9, c8)
        a10 = r"ڏ|ډ|ﺪ|ﺩ"
        b10 = r"د"
        c10 = sub(a10, b10, c9)
        a11 = r"ﺫ|ﺬ|ﻧ"
        b11 = r"ذ"
        c11 = sub(a11, b11, c10)
        a12 = r"ڙ|ڗ|ڒ|ڑ|ڕ|ﺭ|ﺮ"
        b12 = r"ر"
        c12 = sub(a12, b12, c11)
        a13 = r"ﺰ|ﺯ"
        b13 = r"ز"
        c13 = sub(a13, b13, c12)
        a14 = r"ﮊ"
        b14 = r"ژ"
        c14 = sub(a14, b14, c13)
        a15 = r"ݭ|ݜ|ﺱ|ﺲ|ښ|ﺴ|ﺳ"
        b15 = r"س"
        c15 = sub(a15, b15, c14)
        a16 = r"ﺵ|ﺶ|ﺸ|ﺷ"
        b16 = r"ش"
        c16 = sub(a16, b16, c15)
        a17 = r"ﺺ|ﺼ|ﺻ"
        b17 = r"ص"
        c17 = sub(a17, b17, c16)
        a18 = r"ﺽ|ﺾ|ﺿ|ﻀ"
        b18 = r"ض"
        c18 = sub(a18, b18, c17)
        a19 = r"ﻁ|ﻂ|ﻃ|ﻄ"
        b19 = r"ط"
        c19 = sub(a19, b19, c18)
        a20 = r"ﻆ|ﻇ|ﻈ"
        b20 = r"ظ"
        c20 = sub(a20, b20, c19)
        a21 = r"ڠ|ﻉ|ﻊ|ﻋ"
        b21 = r"ع"
        c21 = sub(a21, b21, c20)
        a22 = r"ﻎ|ۼ|ﻍ|ﻐ|ﻏ"
        b22 = r"غ"
        c22 = sub(a22, b22, c21)
        a23 = r"ﻒ|ﻑ|ﻔ|ﻓ"
        b23 = r"ف"
        c23 = sub(a23, b23, c22)
        a24 = r"ﻕ|ڤ|ﻖ|ﻗ"
        b24 = r"ق"
        c24 = sub(a24, b24, c23)
        a25 = r"ڭ|ﻚ|ﮎ|ﻜ|ﮏ|ګ|ﻛ|ﮑ|ﮐ|ڪ|ك"
        b25 = r"ک"
        c25 = sub(a25, b25, c24)
        a26 = r"ﮚ|ﮒ|ﮓ|ﮕ|ﮔ"
        b26 = r"گ"
        c26 = sub(a26, b26, c25)
        a27 = r"ﻝ|ﻞ|ﻠ|ڵ"
        b27 = r"ل"
        c27 = sub(a27, b27, c26)
        a28 = r"ﻡ|ﻤ|ﻢ|ﻣ"
        b28 = r"م"
        c28 = sub(a28, b28, c27)
        a29 = r"ڼ|ﻦ|ﻥ|ﻨ"
        b29 = r"ن"
        c29 = sub(a29, b29, c28)
        a30 = r"ވ|ﯙ|ۈ|ۋ|ﺆ|ۊ|ۇ|ۏ|ۅ|ۉ|ﻭ|ﻮ|ؤ"
        b30 = r"و"
        c30 = sub(a30, b30, c29)
        a31 = r"ﺔ|ﻬ|ھ|ﻩ|ﻫ|ﻪ|ۀ|ە|ة|ہ"
        b31 = r"ه"
        c31 = sub(a31, b31, c30)
        a32 = r"ﭛ|ﻯ|ۍ|ﻰ|ﻱ|ﻲ|ں|ﻳ|ﻴ|ﯼ|ې|ﯽ|ﯾ|ﯿ|ێ|ے|ى|ي"
        b32 = r"ی"
        c32 = sub(a32, b32, c31)
        a33 = r'¬'
        b33 = r'‌'
        c33 = sub(a33, b33, c32)
        pa0 = r'•|·|●|·|・|∙|｡|ⴰ'
        pb0 = r'.'
        pc0 = sub(pa0, pb0, c33)
        pa1 = r',|٬|٫|‚|，'
        pb1 = r'،'
        pc1 = sub(pa1, pb1, pc0)
        pa2 = r'ʕ'
        pb2 = r'؟'
        pc2 = sub(pa2, pb2, pc1)
        na0 = r'۰|٠'
        nb0 = r'0'
        nc0 = sub(na0, nb0, pc2)
        na1 = r'۱|١'
        nb1 = r'1'
        nc1 = sub(na1, nb1, nc0)
        na2 = r'۲|٢'
        nb2 = r'2'
        nc2 = sub(na2, nb2, nc1)
        na3 = r'۳|٣'
        nb3 = r'3'
        nc3 = sub(na3, nb3, nc2)
        na4 = r'۴|٤'
        nb4 = r'4'
        nc4 = sub(na4, nb4, nc3)
        na5 = r'۵'
        nb5 = r'۵'
        nc5 = sub(na5, nb5, nc4)
        na6 = r'۶|٦'
        nb6 = r'6'
        nc6 = sub(na6, nb6, nc5)
        na7 = r'۷|٧'
        nb7 = r'7'
        nc7 = sub(na7, nb7, nc6)
        na8 = r'۸|٨'
        nb8 = r'8'
        nc8 = sub(na8, nb8, nc7)
        na9 = r'۹|٩'
        nb9 = r'9'
        nc9 = sub(na9, nb9, nc8)
        ea1 = r'ـ|ِ|ُ|َ|ٍ|ٌ|ً|'
        eb1 = r''
        ec1 = sub(ea1, eb1, nc9)
        Sa1 = r'( )+'
        Sb1 = r' '
        Sc1 = sub(Sa1, Sb1, ec1)
        Sa2 = r'(\n)+'
        Sb2 = r'\n'
        Sc2 = sub(Sa2, Sb2, Sc1)

        return Sc2

    def space_correction(self, doc_string):
        a00 = r'^(بی|می|نمی)( )'
        b00 = r'\1‌'
        c00 = sub(a00, b00, doc_string)
        a0 = r'( )(می|نمی|بی)( )'
        b0 = r'\1\2‌'
        c0 = sub(a0, b0, c00)
        a1 = r'( )(هایی|ها|های|ایی|هایم|هایت|هایش|هایمان|هایتان|هایشان|ات|ان|ین' \
             r'|انی|بان|ام|ای|یم|ید|اید|اند|بودم|بودی|بود|بودیم|بودید|بودند|ست)( )'
        b1 = r'‌\2\3'
        c1 = sub(a1, b1, c0)
        a2 = r'( )(شده|نشده)( )'
        b2 = r'‌\2‌'
        c2 = sub(a2, b2, c1)
        a3 = r'( )(طلبان|طلب|گرایی|گرایان|شناس|شناسی|گذاری|گذار|گذاران|شناسان|گیری|پذیری|بندی|آوری|سازی|' \
             r'بندی|کننده|کنندگان|گیری|پرداز|پردازی|پردازان|آمیز|سنجی|ریزی|داری|دهنده|آمیز|پذیری' \
             r'|پذیر|پذیران|گر|ریز|ریزی|رسانی|یاب|یابی|گانه|گانه‌ای|انگاری|گا|بند|رسانی|دهندگان|دار)( )'
        b3 = r'‌\2\3'
        c3 = sub(a3, b3, c2)
        return c3

    def space_correction_plus1(self, doc_string):
        out_sentences = ''
        for wrd in doc_string.split(' '):
            try:
                out_sentences = out_sentences + ' ' + self.dic1[wrd]
            except KeyError:
                out_sentences = out_sentences + ' ' + wrd
        return out_sentences

    def space_correction_plus2(self, doc_string):
        out_sentences = ''
        wrds = doc_string.split(' ')
        L = wrds.__len__()
        if L < 2:
            return doc_string
        cnt = 1
        for i in range(0, L - 1):
            w = wrds[i] + wrds[i + 1]
            try:
                out_sentences = out_sentences + ' ' + self.dic2[w]
                cnt = 0
            except KeyError:
                if cnt == 1:
                    out_sentences = out_sentences + ' ' + wrds[i]
                cnt = 1
        if cnt == 1:
            out_sentences = out_sentences + ' ' + wrds[i + 1]
        return out_sentences

    def space_correction_plus3(self, doc_string):
        # Dict = {'گفتوگو': 'گفت‌وگو'}
        out_sentences = ''
        wrds = doc_string.split(' ')
        L = wrds.__len__()
        if L < 3:
            return doc_string
        cnt = 1
        cnt2 = 0
        for i in range(0, L - 2):
            w = wrds[i] + wrds[i + 1] + wrds[i + 2]
            try:
                out_sentences = out_sentences + ' ' + self.dic3[w]
                cnt = 0
                cnt2 = 2
            except KeyError:
                if cnt == 1 and cnt2 == 0:
                    out_sentences = out_sentences + ' ' + wrds[i]
                else:
                    cnt2 -= 1
                cnt = 1
        if cnt == 1 and cnt2 == 0:
            out_sentences = out_sentences + ' ' + wrds[i + 1] + ' ' + wrds[i + 2]
        elif cnt == 1 and cnt2 == 1:
            out_sentences = out_sentences + ' ' + wrds[i + 2]
        return out_sentences

    def normalize(self, doc_string, new_line_elimination=False):


        normalized_string = self.sub_alphabets(doc_string)

        # normalized_string = self.data_helper.clean_text(normalized_string, new_line_elimination).strip()

        normalized_string = re.sub(r"\u200c", " ", normalized_string)  # for normalization

        # if self.statistical_space_correction:
        #     token_list = normalized_string.strip().split()
        #     token_list = [x.strip("\u200c") for x in token_list if len(x.strip("\u200c")) != 0]
        #     token_list = self.token_merger.merg_tokens(token_list, self.token_merger_model, self.half_space_char)
        #     normalized_string = " ".join(x for x in token_list)
        #     normalized_string = self.data_helper.clean_text(normalized_string, new_line_elimination)
        #
        # else:
        #     normalized_string = self.space_correction(self.space_correction_plus1(
        #         self.space_correction_plus2(self.space_correction_plus3(normalized_string)))).strip()
        #

        if self.date_normalizing_needed:
            normalized_string = self.date_normalizer.find_date_part(text_line=normalized_string)

        if self.time_normalizing_needed:
            normalized_string = self.time_normalizer.find_time_part(text_line=normalized_string)

        if self.number_normalizing_needed:
            normalized_string = self.number_normalizer.find_number_part(text_line=normalized_string)

        if self.abbreviation_needed:
            normalized_string = self.abbreviation_normalizer.find_abbreviation_part(text_line=normalized_string)

        if self.pinglish_conversion_needed:
            normalized_string = self.pinglish_conversion.pingilish2persian(
                self.tokenizer.tokenize_words(normalized_string))

        if self.punctuation_normalizing_needed:
            normalized_string = self.punctuation_normalizer.convert_punctuation(text_line=normalized_string)

        if self.space_normalizing_needed:
            normalized_string = self.space_normalizer.convert_space(text_line=normalized_string)

        if self.upper_case:
            return normalized_string.upper()

        return normalized_string


class Spacenormalizer():
    def __init__(self):
        pass
        # self.stemmer = mofid_normalizer.stemmer.FindStems()

    def convert_space(self, text_line):
        pass
        # text_line_list = text_line.split()
        #
        # text = ' '
        # for word in text_line_list:
        #     result = self.stemmer.convert_to_stem(word)
        #     text += result + ' '
        #
        # text = re.sub(' +', ' ', text)
        # return "".join(text.rstrip().lstrip())


class Punctuation():
    def __init__(self):
        self.punctuation_list = []

    def convert_punctuation(self, text_line):
        input_data = re.sub("%|٪", 'درصد', text_line)
        input_data = re.sub(r'[^\w\s]', ' ', input_data)
        input_data = re.findall(r'[\u0600-\u065F\u066A-\u06EF\u06FA-\u06FF].*', input_data)

        return ''.join(input_data)

    def persian_converter(self, input_data):
        print(input_data)
        return input_data


class AbbreviationNormalizer():
    def __init__(self):
        self.abbreviation_rules = r'( ه ق )|( ه ش )|( ه.ق )|( ه.ش )|( ع )|( ره )|(\(ه ش\))|(\(ه ق\))|(\(ه.ش\))|(\(ه.ق\))|(\(ره\))|(\(ع\))|( ه ق )|( ه ش )|( ه.ق. )|( ه.ش. )|( ره )|(\(ه ش\))|(\(ه ق\))|(\(.ه.ش\))|(\(ه.ق.\))|(\(ره\))|(\(ع\))'
        self.abbreviation_dictionary = {
            'هق': ' هجری قمری ',
            'هش': ' هجری شمسی ',
            'ره': ' رحمه الله ',
            'ع': ' علیه السلام '
        }

    def find_abbreviation_part(self, text_line):
        content_new = re.sub(self.abbreviation_rules, lambda x: self.abbreviation_converter(x.group()), text_line)

        return content_new

    def abbreviation_converter(self, input_data):
        try:
            return self.abbreviation_dictionary[re.sub('\W+', "", input_data)]
        except:
            return input_data


class TimeNormalizer():
    def __init__(self):

        self.time_rules = r'(([0-1]?[0-9]|2[0-3]):[0-5]?[0-9](:[0-5]?[0-9])?)|(([0-5]?[0-9]):([0-5]?[0-9]) )'
        self.time_format = ['%H:%M:%S', '%H:%M', '%M:%S']
        self.minute_dictionary = {15: "ربع", 30: "نیم"}

    def find_time_part(self, text_line=""):

        content_new = re.sub(self.time_rules, lambda x: self.time_digit2string(x.group()), text_line)

        return content_new

    def minute_converter(self, time):
        # ranodmly convert 30,15 to نیم /ربع

        rnd = random.choices([1, 2], weights=(0.6, 0.4), k=1)
        if rnd[0] == 2 and time.second == 0:
            return self.minute_dictionary[time.minute]

        return words(time.minute) + " دقیقه"

    def time_digit2string(self, input_time):

        for form in self.time_format:
            try:
                d = datetime.strptime(input_time, form)

                hour = words(d.hour) + " و "
                minute = words(d.minute) + " دقیقه" if d.minute not in [15, 30] else self.minute_converter(d)
                second = "و " + words(d.second) + " ثانیه " if str(words(d.second)) != 'صفر' else ''

                return hour + " " + minute + " " + second

            except ValueError:

                continue

        return input_time


class NumberNormalizer():
    def __init__(self):
        self.number_rules = r'( +| -)?(((\d{1,30})?[./]\d{1,30})|(\d{1,20}))'

    def find_number_part(self, text_line):

        content_new = re.sub(self.number_rules, lambda x: self.number_converter((x.group())), text_line)

        return content_new

    def number_converter(self, input_data):

        try:
            return " " + words(input_data, positive='', negative='منهای ', fraction_separator=' تقسیم بر ',
                               ordinal_denominator=False, decimal_separator=' ممیز ') + "  "

        except ValueError:
            return input_data


class DateNormalizer():
    def __init__(self):

        self.month_names = {'شمسی': {1: "فروردین", 2: "اردیبهشت", 3: "خرداد", 4:
            "تیر", 5: "مرداد", 6: "شهریور", 7:
                                         "مهر", 8: "آبان", 9: "آذر", 10:
                                         "دی", 11: "بهمن", 12: "اسفند"},

                            'قمری': {1: "محرم", 2: "صفر", 3: "ربیع‌الاول", 4:
                                "ربیع‌الثانی", 5: "جمادی‌الاول", 6: "جمادی‌الثانی", 7:
                                         "رجب", 8: "شعبان", 9: "رمضان", 10:
                                         "شوال", 11: "ذیقعده", 12: "ذیحجه"},

                            'میلادی': {1: "ژانویه", 2: "فوریه", 3: "مارچ", 4:
                                "آپریل", 5: "می", 6: "جون", 7:
                                           "جولای", 8: "آگوست", 9: "سپتامبر", 10:
                                           "اکتبر", 11: "نوامبر", 12: "دسامبر"}}

        self.num_dict = {"صد": 100, "هزار": 1000, "میلیون": 1000000, "دویست": 200,
                         "ده": 10, "نه": 9, "هشت": 8, "هفت": 7, "شش": 6, "پنج": 5,
                         "چهار": 4, "سه": 3, "دو": 2, "یک": 1, "یازده": 11, "سیزده": 13,
                         "چهارده": 14, "دوازده": 12, "پانزده": 15, "شانزده": 16, "هفده": 17,
                         "هجده": 18, "نوزده": 19, "بیست": 20, "سی": 30, "چهل": 40, "پنجاه": 50,
                         "شصت": 60, "هفتاد": 70, "نود": 90, "سیصد": 300, "چهارصد": 400,
                         "پانصد": 500, "ششصد": 600, "هفتصد": 700, "هشتصد": 800, "نهصد": 900,
                         "هشتاد": 80, " ": 0, "میلیارد": 1000000000,
                         "صدم": 100, "هزارم": 1000, "دویستم": 200,
                         "دهم": 10, "نهم": 9, "هشتم": 8, "هفتم": 7, "ششم": 6, "پنجم": 5,
                         "چهارم": 4, "سوم": 3, "دوم": 2, "یکم": 1, "اول": 1, "یازدهم": 11, "سیزدهم": 13,
                         "چهاردهم": 14, "دوازدهم": 12, "پانزدهم": 15, "شانزدهم": 16, "هفدهم": 17,
                         "هجدهم": 18, "نوزدهم": 19, "بیستم": 20, "چهلم": 40, "پنجاهم": 50,
                         "شصتم": 60, "هفتادم": 70, "نودم": 90, "سیصدم": 300, "چهارصدم": 400,
                         "پانصدم": 500, "ششصدم": 600, "هفتصدم": 700, "هشتصدم": 800, "نهصدم": 900,
                         "هشتادم": 80}

        self.format_list = ['%Y-%m-%d', '%Y-%d-%m', '%d-%m-%Y', '%m-%d-%Y', '%Y/%m/%d', '%Y/%d/%m', '%d/%m/%Y',
                            '%m/%d/%Y']

        self.date_rules = r'(((0[1-9]|1[012]|\b[1-9])[-/.](0[1-9]|[12][0-9]|3[01]|[1-9])[-/.][11-20]\d\d\d)|((0[0-9]|[12][0-9]|3[01]|\b[1-9])[-/.](0[1-9]|1[012]|[1-9])[-/.]([11-20]\d\d\d))|(([11-20]\d\d\d)[-/.](0[1-9]|1[012]|[1-9])[-/.](0[1-9]|[12][0-9]|3[01]|[1-9]\b))|(([11-20]\d\d\d)[-/.](0[1-9]|[12][0-9]|3[01]|[1-9])[-/.](0[1-9]|1[012]|[1-9]\b)))'

    def find_date_part(self, text_line=""):

        content_new = re.sub(self.date_rules, lambda x: self.date_digit2string(x.group()), text_line)

        return content_new

    def convert_month(self, month, date_type):
        rnd = random.choices([1, 2, 3], weights=(0.6, 0.3, 0.1), k=1)

        if rnd[0] == 1: return self.month_names[date_type][month]
        if rnd[0] == 2: return self.month_names[date_type][month] + " ماه "
        if rnd[0] == 3: return str(words(month))

    def convert_year(self, year, date_type):

        rnd = random.choices([1, 2, 3, 4], weights=(0.4, 0.2, 0.1, 0.2), k=1)

        if rnd[0] == 1: return words(year)
        if rnd[0] == 2: return "سال " + words(year)
        if rnd[0] == 3: return "سال " + words(year) + ' ' + date_type
        if rnd[0] == 4: return words(year) + ' ' + date_type

    def convert_day(self, day):
        rnd = random.choices([1, 2], weights=(0.4, 0.6), k=1)

        if rnd[0] == 1: return words(day)
        if rnd[0] == 2: return ordinal_words(day)

    def date_type(self, date):
        if int(date) < 1410:
            return 'شمسی'
        elif 1550 > int(date) > 1410:
            return 'قمری'
        else:
            return 'میلادی'

    def date_digit2string(self, input_date):

        split_date = re.split("-|/|\.", input_date)
        if 1000 < int(split_date[0]) < 3000 and 0 < int(split_date[1]) < 13 and 0 < int(split_date[2]) < 32:

            date_type = self.date_type(int(split_date[0]))

            month = self.convert_month(int(split_date[1]), date_type)
            year = self.convert_year(int(split_date[0]), date_type)
            day = self.convert_day(int(split_date[2]))

            return day + " " + month + " " + year

        elif 1000 < int(split_date[0]) < 3000 and 0 < int(split_date[2]) < 13 and 0 < int(split_date[1]) < 32:

            date_type = self.date_type(int(split_date[0]))

            month = self.convert_month(int(split_date[2]), date_type)
            year = self.convert_year(int(split_date[0]), date_type)
            day = self.convert_day(int(split_date[1]))
            return day + " " + month + " " + year

        elif 1000 < int(split_date[2]) < 3000 and 0 < int(split_date[1]) < 13 and 0 < int(split_date[0]) < 32:

            date_type = self.date_type(int(split_date[2]))

            month = self.convert_month(int(split_date[1]), date_type)
            year = self.convert_year(int(split_date[2]), date_type)
            day = self.convert_day(int(split_date[0]))

            return day + " " + month + " " + year

        elif 1000 < int(split_date[2]) < 3000 and 0 < int(split_date[0]) < 13 and 0 < int(split_date[1]) < 32:

            date_type = self.date_type(int(split_date[2]))

            month = self.convert_month(int(split_date[0]), date_type)
            year = self.convert_year(int(split_date[2]), date_type)
            day = self.convert_day(int(split_date[1]))

            return day + " " + month + " " + year

        else:

            return input_date


class PinglishNormalizer():
    def __init__(self):
        self.data_helper = DataHelper()
        self.file_dir = os.path.dirname(os.path.realpath(__file__)) + "/"

        self.en_dict_filename = self.file_dir + "resource/tokenizer/enDict"
        self.en_dict = self.data_helper.load_var(self.en_dict_filename)

        self.fa_dict_filename = self.file_dir + "resource/tokenizer/faDict"
        self.fa_dict = self.data_helper.load_var(self.fa_dict_filename)

    def pingilish2persian(self, pinglish_words_list):

        for i, word in enumerate(pinglish_words_list):
            if word in self.en_dict:
                pinglish_words_list[i] = self.en_dict[word]  # .decode("utf-8")
                # inp = inp.replace(word, enDict[word], 1)
            else:
                ch = self.characterize(word)
                pr = self.map_char(ch)
                amir = self.make_word(pr)
                for wd in amir:
                    am = self.escalation(wd)
                    asd = ''.join(am)
                    if asd in self.fa_dict:
                        pinglish_words_list[i] = asd  # .decode("utf-8")
                        # inp = inp.replace(word, asd, 1)
        inp = " ".join(x for x in pinglish_words_list)
        return inp

    def characterize(self, word):
        list_of_char = []
        i = 0
        while i < len(word):
            char = word[i]
            sw_out = self.switcher(char)
            if (sw_out == None):
                esp_out = None
                if (i < len(word) - 1):
                    esp_out = self.esp_check(word[i], word[i + 1])
                if (esp_out == None):
                    list_of_char.append(word[i])
                else:
                    list_of_char.append(esp_out)
                    i += 1
            else:
                list_of_char.append(sw_out)
            i += 1
        return list_of_char

    def switcher(self, ch):
        switcher = {
            "c": None,
            "k": None,
            "z": None,
            "s": None,
            "g": None,
            "a": None,
            "u": None,
            "e": None,
            "o": None
        }
        return switcher.get(ch, ch)

    def esp_check(self, char1, char2):
        st = char1 + char2
        if (st == "ch"):
            return "ch"
        elif (st == "kh"):
            return "kh"
        elif (st == "zh"):
            return "zh"
        elif (st == "sh"):
            return "sh"
        elif (st == "gh"):
            return "gh"
        elif (st == "aa"):
            return "aa"
        elif (st == "ee"):
            return "ee"
        elif (st == "oo"):
            return "oo"
        elif (st == "ou"):
            return "ou"
        else:
            return None

    def map_char(self, word):
        listm = []
        sw_out = self.map_switcher(word[0])
        i = 0
        if (sw_out == None):
            listm.append(["ا"])
            i += 1
        if (word[0] == "oo"):
            listm.append(["او"])
            i += 1
        while i < len(word):
            listm.append(self.char_switcher(word[i]))
            i += 1
        if word[len(word) - 1] == "e":
            listm.append(["ه"])
        elif word[len(word) - 1] == "a":
            listm.append(["ا"])
        elif word[len(word) - 1] == "o":
            listm.append(["و"])
        elif word[len(word) - 1] == "u":
            listm.append(["و"])

        return listm

    def map_switcher(self, ch):
        switcher = {
            "a": None,
            "e": None,
            "o": None,
            "u": None,
            "ee": None,

            "ou": None
        }
        return switcher.get(ch, ch)

    def make_word(self, chp):
        word_list = [[]]
        for char in chp:
            word_list_temp = []
            for tmp_word_list in word_list:
                for chch in char:
                    tmp = copy.deepcopy(tmp_word_list)
                    tmp.append(chch)
                    word_list_temp.append(tmp)
            word_list = word_list_temp
        return word_list

    def escalation(self, word):
        tmp = []
        i = 0
        t = len(word)
        while i < t - 1:
            tmp.append(word[i])
            if word[i] == word[i + 1]:
                i += 1
            i += 1
        if i != t:
            tmp.append(word[i])
        return tmp

    def char_switcher(self, ch):
        switcher = {
            'a': ["", "ا"],
            'c': ["ث", "ص", "ص"],
            'h': ["ه", "ح"],
            'b': ["ب"],
            'p': ["پ"],
            't': ["ت", "ط"],
            's': ["س", "ص", "ث"],
            'j': ["ج"],
            'ch': ["چ"],
            'kh': ["خ"],
            'q': ["ق", "غ"],
            'd': ["د"],
            'z': ["ز", "ذ", "ض", "ظ"],
            'r': ["ر"],
            'zh': ["ژ"],
            'sh': ["ش"],
            'gh': [",ق", "غ"],
            'f': ["ف"],
            'k': ["ک"],
            'g': ["گ"],
            'l': ["ل"],
            'm': ["م"],
            'n': ["ن"],
            'v': ["و"],
            'aa': ["ا"],
            'ee': ["ی"],
            'oo': ["و"],
            'ou': ["و"],
            'i': ["ی"],
            'y': ["ی"],
            ' ': [""],
            'w': ["و"],
            'e': ["", "ه"],
            'o': ["", "و"]
        }
        return switcher.get(ch, "")



from num2fawords import words, ordinal_words
import re

import random


class Date2String():

    def __init__(self):

        self.month_names = {'شمسی': {1: "فروردین", 2: "اردیبهشت", 3: "خرداد", 4:
            "تیر", 5: "مرداد", 6: "شهریور", 7:
                                         "مهر", 8: "آبان", 9: "آذر", 10:
                                         "دی", 11: "بهمن", 12: "اسفند"},

                            'قمری': {1: "محرم", 2: "صفر", 3: "ربیع‌الاول", 4:
                                "ربیع‌الثانی", 5: "جمادی‌الاول", 6: "جمادی‌الثانی", 7:
                                         "رجب", 8: "شعبان", 9: "رمضان", 10:
                                         "شوال", 11: "ذیقعده", 12: "ذیحجه"},

                            'میلادی': {1: "ژانویه", 2: "فوریه", 3: "مارچ", 4:
                                "آپریل", 5: "می", 6: "جون", 7:
                                           "جولای", 8: "آگوست", 9: "سپتامبر", 10:
                                           "اکتبر", 11: "نوامبر", 12: "دسامبر"}}

        self.num_dict = {"صد": 100, "هزار": 1000, "میلیون": 1000000, "دویست": 200,
                         "ده": 10, "نه": 9, "هشت": 8, "هفت": 7, "شش": 6, "پنج": 5,
                         "چهار": 4, "سه": 3, "دو": 2, "یک": 1, "یازده": 11, "سیزده": 13,
                         "چهارده": 14, "دوازده": 12, "پانزده": 15, "شانزده": 16, "هفده": 17,
                         "هجده": 18, "نوزده": 19, "بیست": 20, "سی": 30, "چهل": 40, "پنجاه": 50,
                         "شصت": 60, "هفتاد": 70, "نود": 90, "سیصد": 300, "چهارصد": 400,
                         "پانصد": 500, "ششصد": 600, "هفتصد": 700, "هشتصد": 800, "نهصد": 900,
                         "هشتاد": 80, " ": 0, "میلیارد": 1000000000,
                         "صدم": 100, "هزارم": 1000, "دویستم": 200,
                         "دهم": 10, "نهم": 9, "هشتم": 8, "هفتم": 7, "ششم": 6, "پنجم": 5,
                         "چهارم": 4, "سوم": 3, "دوم": 2, "یکم": 1, "اول": 1, "یازدهم": 11, "سیزدهم": 13,
                         "چهاردهم": 14, "دوازدهم": 12, "پانزدهم": 15, "شانزدهم": 16, "هفدهم": 17,
                         "هجدهم": 18, "نوزدهم": 19, "بیستم": 20, "چهلم": 40, "پنجاهم": 50,
                         "شصتم": 60, "هفتادم": 70, "نودم": 90, "سیصدم": 300, "چهارصدم": 400,
                         "پانصدم": 500, "ششصدم": 600, "هفتصدم": 700, "هشتصدم": 800, "نهصدم": 900,
                         "هشتادم": 80}

        self.format_list = ['%Y-%m-%d', '%Y-%d-%m', '%d-%m-%Y', '%m-%d-%Y', '%Y/%m/%d', '%Y/%d/%m', '%d/%m/%Y',
                            '%m/%d/%Y']

        self.date_rules = r'(((0[1-9]|1[012]|\b[1-9])( )*[-/.]( )*(0[1-9]|[12][0-9]|3[01]|[1-9])( )*[-/.]( )*[11-20]\d\d\d)|((0[0-9]|[12][0-9]|3[01]|\b[1-9])( )*[-/.]( )*(0[1-9]|1[012]|[1-9])( )*[-/.]( )*([11-20]\d\d\d))|(([11-20]\d\d\d)( )*[-/.]( )*(0[1-9]|1[012]|[1-9])( )*[-/.]( )*(0[1-9]|[12][0-9]|3[01]|[1-9]\b))|(([11-20]\d\d\d)( )*[-/.]( )*(0[1-9]|[12][0-9]|3[01]|[1-9])( )*[-/.]( )*(0[1-9]|1[012]|[1-9]\b)))'

    def normalize(self, doc_string):

        normalized_string = self.find_date_part(text_line=doc_string)

        return normalized_string

    def find_date_part(self, text_line=""):

        content_new = re.sub(self.date_rules, lambda x: self.date_digit2string(x.group()), text_line)

        return content_new

    def convert_month(self, month, date_type):
        rnd = random.choices([1, 2, 3], weights=(0.6, 0.3, 0.1), k=1)

        if rnd[0] == 1: return self.month_names[date_type][month]
        if rnd[0] == 2: return self.month_names[date_type][month] + " ماه "
        if rnd[0] == 3: return str(words(month))

    def convert_year(self, year, date_type):

        rnd = random.choices([1, 2, 3, 4], weights=(0.4, 0.2, 0.1, 0.2), k=1)

        if rnd[0] == 1: return words(year)
        if rnd[0] == 2: return "سال " + words(year)
        if rnd[0] == 3: return "سال " + words(year) + ' ' + date_type
        if rnd[0] == 4: return words(year) + ' ' + date_type

    def convert_day(self, day):
        rnd = random.choices([1, 2], weights=(0.4, 0.6), k=1)

        if rnd[0] == 1: return words(day)
        if rnd[0] == 2: return ordinal_words(day)

    def date_type(self, date):
        if int(date) < 1410:
            return 'شمسی'
        elif 1550 > int(date) > 1410:
            return 'قمری'
        else:
            return 'میلادی'

    def date_digit2string(self, input_date):

        split_date = re.split("-|/|\.", input_date)
        if 1000 < int(split_date[0]) < 3000 and 0 < int(split_date[1]) < 13 and 0 < int(split_date[2]) < 32:

            date_type = self.date_type(int(split_date[0]))

            month = self.convert_month(int(split_date[1]), date_type)
            year = self.convert_year(int(split_date[0]), date_type)
            day = self.convert_day(int(split_date[2]))

            return day + " " + month + " " + year

        elif 1000 < int(split_date[0]) < 3000 and 0 < int(split_date[2]) < 13 and 0 < int(split_date[1]) < 32:

            date_type = self.date_type(int(split_date[0]))

            month = self.convert_month(int(split_date[2]), date_type)
            year = self.convert_year(int(split_date[0]), date_type)
            day = self.convert_day(int(split_date[1]))
            return day + " " + month + " " + year

        elif 1000 < int(split_date[2]) < 3000 and 0 < int(split_date[1]) < 13 and 0 < int(split_date[0]) < 32:

            date_type = self.date_type(int(split_date[2]))

            month = self.convert_month(int(split_date[1]), date_type)
            year = self.convert_year(int(split_date[2]), date_type)
            day = self.convert_day(int(split_date[0]))

            return day + " " + month + " " + year

        elif 1000 < int(split_date[2]) < 3000 and 0 < int(split_date[0]) < 13 and 0 < int(split_date[1]) < 32:

            date_type = self.date_type(int(split_date[2]))

            month = self.convert_month(int(split_date[0]), date_type)
            year = self.convert_year(int(split_date[2]), date_type)
            day = self.convert_day(int(split_date[1]))

            return day + " " + month + " " + year

        else:

            return input_date



# import spacy
# @spacy.registry.tokenizers("whitespace_tokenizer")def create_whitespace_tokenizer():
#     def create_tokenizer(nlp):
#         return WhitespaceTokenizer(nlp.vocab)
#
#     return create_tokenizer


import pickle
import re
import string
import os
from urllib.parse import urljoin

class DataHelper():
    def __init__(self):
        pass

    def clean_text(self, text_doc, new_line_elimination):
        punctuations = r')(}{:؟!،؛»«.' + r"/<>?.,:;"
        punctuations = '[' + punctuations + string.punctuation + ']'
        punctuations = punctuations.replace("@", "")

        text_doc.strip()

        # pattern = ur'\s*@[a-zA-Z0-9]*\s*'
        # tmp = re.findall(pattern, text_doc)
        # newstring = re.sub(pattern, eliminate_pattern, text_doc)


        #finding the numbers

        # pattern = r"[-+]?\d*\.\d+|\d+"
        # nums_list = re.findall(pattern, text_doc)
        # newstring = re.sub(pattern, 'floatingpointnumber', text_doc)

        # pattern = r'(?i)\b((?:[a-z][\w-]+:(?:/{1,3}|[a-z0-9%])|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:\'".,<>?«»“”‘’]))'
        pattern =r'(?i)\b((?:https?:(?:/{1,3}|[a-z0-9%])|[a-z0-9.\-]+[.](?:com|net|org|edu|gov|mil|aero|asia|biz|cat|coop|info|int|jobs|mobi|museum|name|post|pro|tel|travel|xxx|ac|ad|ae|af|ag|ai|al|am|an|ao|aq|ar|as|at|au|aw|ax|az|ba|bb|bd|be|bf|bg|bh|bi|bj|bm|bn|bo|br|bs|bt|bv|bw|by|bz|ca|cc|cd|cf|cg|ch|ci|ck|cl|cm|cn|co|cr|cs|cu|cv|cx|cy|cz|dd|de|dj|dk|dm|do|dz|ec|ee|eg|eh|er|es|et|eu|fi|fj|fk|fm|fo|fr|ga|gb|gd|ge|gf|gg|gh|gi|gl|gm|gn|gp|gq|gr|gs|gt|gu|gw|gy|hk|hm|hn|hr|ht|hu|id|ie|il|im|in|io|iq|ir|is|it|je|jm|jo|jp|ke|kg|kh|ki|km|kn|kp|kr|kw|ky|kz|la|lb|lc|li|lk|lr|ls|lt|lu|lv|ly|ma|mc|md|me|mg|mh|mk|ml|mm|mn|mo|mp|mq|mr|ms|mt|mu|mv|mw|mx|my|mz|na|nc|ne|nf|ng|ni|nl|no|np|nr|nu|nz|om|pa|pe|pf|pg|ph|pk|pl|pm|pn|pr|ps|pt|pw|py|qa|re|ro|rs|ru|rw|sa|sb|sc|sd|se|sg|sh|si|sj|Ja|sk|sl|sm|sn|so|sr|ss|st|su|sv|sx|sy|sz|tc|td|tf|tg|th|tj|tk|tl|tm|tn|to|tp|tr|tt|tv|tw|tz|ua|ug|uk|us|uy|uz|va|vc|ve|vg|vi|vn|vu|wf|ws|ye|yt|yu|za|zm|zw)/)(?:[^\s()<>{}\[\]]+|\([^\s()]*?\([^\s()]+\)[^\s()]*?\)|\([^\s]+?\))+(?:\([^\s()]*?\([^\s()]+\)[^\s()]*?\)|\([^\s]+?\)|[^\s`!()\[\]{};:\'".,<>?«»“”‘’])|(?:(?<!@)[a-z0-9]+(?:[.\-][a-z0-9]+)*[.](?:com|net|org|edu|gov|mil|aero|asia|biz|cat|coop|info|int|jobs|mobi|museum|name|post|pro|tel|travel|xxx|ac|ad|ae|af|ag|ai|al|am|an|ao|aq|ar|as|at|au|aw|ax|az|ba|bb|bd|be|bf|bg|bh|bi|bj|bm|bn|bo|br|bs|bt|bv|bw|by|bz|ca|cc|cd|cf|cg|ch|ci|ck|cl|cm|cn|co|cr|cs|cu|cv|cx|cy|cz|dd|de|dj|dk|dm|do|dz|ec|ee|eg|eh|er|es|et|eu|fi|fj|fk|fm|fo|fr|ga|gb|gd|ge|gf|gg|gh|gi|gl|gm|gn|gp|gq|gr|gs|gt|gu|gw|gy|hk|hm|hn|hr|ht|hu|id|ie|il|im|in|io|iq|ir|is|it|je|jm|jo|jp|ke|kg|kh|ki|km|kn|kp|kr|kw|ky|kz|la|lb|lc|li|lk|lr|ls|lt|lu|lv|ly|ma|mc|md|me|mg|mh|mk|ml|mm|mn|mo|mp|mq|mr|ms|mt|mu|mv|mw|mx|my|mz|na|nc|ne|nf|ng|ni|nl|no|np|nr|nu|nz|om|pa|pe|pf|pg|ph|pk|pl|pm|pn|pr|ps|pt|pw|py|qa|re|ro|rs|ru|rw|sa|sb|sc|sd|se|sg|sh|si|sj|Ja|sk|sl|sm|sn|so|sr|ss|st|su|sv|sx|sy|sz|tc|td|tf|tg|th|tj|tk|tl|tm|tn|to|tp|tr|tt|tv|tw|tz|ua|ug|uk|us|uy|uz|va|vc|ve|vg|vi|vn|vu|wf|ws|ye|yt|yu|za|zm|zw)\b/?(?!@)))'

        newstring = re.sub(pattern,'',text_doc)

        pattern = '\s*' + punctuations + '+' + '\s*'
        tmp = re.findall(pattern, newstring)
        newstring = re.sub(pattern, self.add_space, newstring)

        # pattern = u'([a-zA-Z0-9]+)(\s*)(' + punctuations + u')(\s*)([a-zA-Z0-9]+)'
        # rep = ur'\1\3\5'
        # tmp = re.findall(pattern, newstring)
        # newstring = re.sub(pattern, rep, newstring)

        pattern = r'[\n]+'
        tmp = re.findall(pattern, newstring)
        if new_line_elimination:
            newstring = re.sub(pattern, " ", newstring)
        else:
            # newstring = re.sub(pattern, "\n", newstring)
            pass

        punctuations = r")(}{:؟!-،؛»«.@$&%" + r"/<>?.,:;"
        latinLettersDigits = r"a-zA-Z0-9"
        pattern = r'[^' + punctuations + latinLettersDigits + 'آ-ی' + '‌' + '\d\s:]'
        tmp = re.findall(pattern, newstring)
        newstring = re.sub(pattern, self.eliminate_pattern, newstring)

        pattern = r'[ ]+'
        tmp = re.findall(pattern, newstring)
        newstring = re.sub(pattern, ' ', newstring)

        #
        # for number in nums_list:
        #     pattern = 'floatingpointnumber'
        #     newstring = re.sub(pattern, number, newstring, 1)


        return newstring



    def replace_newline_with_dot(self, mystring):
        return ' . '

    def eliminate_pattern(self, mystring):
        return ""

    def load_var(self, load_path):
        file = open(load_path, 'rb')
        variable = pickle.load(file)
        file.close()
        return variable

    def save_var(self, save_path, variable):
        print("saving vars ...")
        file = open(save_path, 'wb')
        pickle.dump(variable, file)
        print("variable saved.")
        file.close()

    def build_stem_dictionary(self, normalizer, verb_tense_path, mokasar_noun_path):
        path_dir = "resource/Persian_Dependency_Treebank/Data/2ndRep"
        lexicon_stem = set()
        verb_stem = set()
        #verb_tense_map = {}
        verb_p2f_map = {}
        verb_f2p_map = {}
        for fileName in os.listdir(path_dir):
            file_path = path_dir + "/" + fileName
            with open(file_path, "r") as input:
                input_content = input.readlines()
                for el in input_content:
                    el = normalizer.sub_alphabets(el)
                    el = el.split("\t")
                    if (len(el) > 2):
                        if (el[3] == 'V'):
                            tmp_pos = "V"
                        else:
                            tmp_pos = "N"
                        stem_word = el[2]
                        stem_word = stem_word.split("#")
                        stem_word = [x.strip('\u200c') for x in stem_word]
                        if (tmp_pos == "V" and len(stem_word) == 2):
                            if (len(stem_word[0]) != 0 and len(stem_word[1]) != 0):
                                verb_p2f_map[stem_word[0]] = stem_word[1]
                                verb_f2p_map[stem_word[1]] = stem_word[0]
                                verb_stem.add(stem_word[0])
                                verb_stem.add(stem_word[1])
                        if(tmp_pos == 'V' and len(stem_word) == 3):
                            if(len(stem_word[0]) != 0 and len(stem_word[1]) != 0 and len(stem_word[2]) !=0):
                                #verb_prifix.add(stem_word[0])
                                verb_p2f_map[stem_word[1]] = stem_word[2]
                                verb_f2p_map[stem_word[2]] = stem_word[1]
                                verb_stem.add(stem_word[1])
                                verb_stem.add(stem_word[2])
                        for t in stem_word:
                            if len(t) > 1:
                                if (tmp_pos == 'N'):
                                    lexicon_stem.add(t)

        with open(verb_tense_path, "r") as bon_file:
            bon_file_content = bon_file.readlines()
            for el in bon_file_content:
                el = el.strip()
                el = normalizer.sub_alphabets(el)
                el = el.split()
                el = [x.strip('\u200c') for x in el]

                verb_p2f_map[el[0]] = el[1]
                verb_f2p_map[el[1]] = el[0]
                verb_stem.add(el[0])
                verb_stem.add(el[1])

        irregular_noun = {}
        with open(mokasar_noun_path, "r") as input:
            input_content = input.readlines()
            for el in input_content:
                el = normalizer.sub_alphabets(el)
                el = el.replace("\t\t", "\t")
                el = el.strip().split("\t")
                el = [x.strip('\u200c') for x in el]
                irregular_noun[el[0]] = el[1]
                lexicon_stem.add(el[0])

        verb_tense_map = [verb_p2f_map, verb_f2p_map]
        return lexicon_stem, verb_stem, verb_tense_map, irregular_noun



from .chunker import FindChunks
from .data_helper import DataHelper
from .dependency import DependencyParser
from .normalizer import Normalizer
from .postagger import POSTagger
from .stemmer import FindStems
from .tokenizer import Tokenizer
from .token_merger import ClassifierChunkParser
# from .spell_checker import SpellCheck
from .char_normalizer import Char_Normalizer
from .date2string import Date2String
from .time2string import Time2String
from .num2string import Num2String
from .abbreviation import Abbreviation
from .punctuation_remover import Punc_remover
from .affix_norm import AffixNorm
from .stemmer import FindStems
from .mapping_words import Word_Mapper


from num2fawords import words, ordinal_words
import re
from datetime import datetime
import random


class Time2String():

    def __init__(self):

        self.time_rules = r'( ([0-1]?[0-9]|2[0-3])( +)?:( +)?[0-5]?[0-9](( +)?:( +)?[0-5]?[0-9])?) | (([0-5]?[0-9])( +)?:( +)?([0-5]?[0-9]) )'
        self.time_format = ['%H:%M:%S', '%H:%M', '%M:%S']
        self.minute_dictionary = {15: "ربع", 30: "نیم"}

    def normalize(self, doc_string):

        normalized_string = self.find_time_part(text_line=doc_string)

        normalized_string = re.sub(r" +", " ", normalized_string)

        return normalized_string

    def find_time_part(self, text_line=""):

        content_new = re.sub(self.time_rules, lambda x: self.time_digit2string(x.group()), " " + text_line + " ")

        return content_new

    def minute_converter(self, time):
        # ranodmly convert 30,15 to نیم /ربع

        rnd = random.choices([1, 2], weights=(0.6, 0.4), k=1)
        if rnd[0] == 2 and time.second == 0:
            return self.minute_dictionary[time.minute]

        return words(time.minute) + " دقیقه"

    def time_digit2string(self, input_time):

        input_time = re.sub(" ", "", input_time)

        for form in self.time_format:
            try:
                d = datetime.strptime(input_time, form)

                hour = words(d.hour) + " و " if words(d.hour) != "صفر" else ""
                minute = words(d.minute) + " دقیقه" if d.minute not in [15, 30] else self.minute_converter(d)
                second = "و " + words(d.second) + " ثانیه " if str(words(d.second)) != 'صفر' else ''

                return " " + hour + " " + minute + " " + second

            except ValueError:

                continue

        return " " + input_time + " "



import re


class Abbreviation():

    def __init__(self):

        self.abbreviation_rules = r'( ه ق )|( ه ش )|( ه.ق )|( ه.ش )|( ع )|( ره )|(\(ه ش\))|(\(ه ق\))|(\(ه.ش\))|(\(ه.ق\))|(\(ره\))|(\(ع\))|( ه ق )|( ه ش )|( ه.ق. )|( ه.ش. )|( ره )|(\(ه ش\))|(\(ه ق\))|(\(.ه.ش\))|(\(ه.ق.\))|(\(ره\))|(\(ع\))'
        self.abbreviation_dictionary = {
            'هق': ' هجری قمری ',
            'هش': ' هجری شمسی ',
            'ره': ' رحمه الله ',
            'ع': ' علیه السلام '
        }

    def normalize(self, doc_string):
        normalized_string = self.find_abbreviation_part(text_line=doc_string)

        return normalized_string

    def find_abbreviation_part(self, text_line):
        content_new = re.sub(self.abbreviation_rules, lambda x: self.abbreviation_converter(x.group()), " "+text_line+" ")

        return content_new.lstrip().strip()

    def abbreviation_converter(self, input_data):
        try:
            return self.abbreviation_dictionary[re.sub('\W+', "", input_data)]
        except:
            return input_data






import os
from nltk.tag.stanford import StanfordPOSTagger
import re

class POSTagger():
    def __init__(self,
                 stanford_postagger_model=None,
                 wapiti_postagger_model=None,
                 jar_tagger_path=None,
                 jdk_variable_path="C:/Program Files/Java/jdk1.8.0_121/bin/java.exe",
                 tagging_model="wapiti"):

        import platform
        if platform.system() == "Windows":
            self.tagging_model = "stanford"
        else:
            self.tagging_model = tagging_model

        self.dir_path = os.path.dirname(os.path.realpath(__file__)) + "/"

        if stanford_postagger_model is None:
            self.stanford_postagger_model = self.dir_path + "resource/postagger/NC_model"
        else:
            self.stanford_postagger_model = stanford_postagger_model

        if jar_tagger_path is None:
            self.jar_tagger_path = self.dir_path + 'resource/postagger/stanford-postagger.jar'
        else:
            self.jar_tagger_path = jar_tagger_path

        if wapiti_postagger_model is None:
            self.wapiti_postagger_model = self.dir_path + "resource/postagger/UPC_full_model_wapiti"
        else:
            self.wapiti_postagger_model = wapiti_postagger_model

        if self.tagging_model == "stanford":
            java_path = jdk_variable_path
            os.environ['JAVAHOME'] = java_path

            self.tagger = StanfordPOSTagger(model_filename=self.stanford_postagger_model,
                                            path_to_jar=self.jar_tagger_path,
                                            encoding='utf-8',
                                            java_options='-mx5000m')
        elif self.tagging_model == "wapiti":
            from wapiti import Model
            self.tagger = Model(model=self.wapiti_postagger_model)

    def is_all_latin(self, word):
        pattern = '[a-zA-Z]*'
        w = re.sub(pattern, '', word)
        if len(w) == 0:
            return True
        else:
            return False

    def parse(self, token_list):
        tagged_tuples = []
        if self.tagging_model == "stanford":
            postags = self.tagger.tag(token_list)
            for element in postags:
                tmp = '_'.join(t for t in element)
                tmp = tmp.strip("_")
                tmp = tmp.split('/')
                tag = tmp[-1]
                tmp = tmp[:-1]
                tmp = '/'.join(i for i in tmp)
                tmp = tmp.strip('/')
                if self.is_all_latin(tmp):
                    tagged_tuples.append((tmp, "FW"))
                else:
                    tagged_tuples.append((tmp, tag))

        elif self.tagging_model == "wapiti":
            sent_line = "\n".join(x for x in token_list)
            postags = self.tagger.label_sequence(sent_line).decode('utf-8')
            postags = postags.strip().split('\n')
            for i, el in enumerate(token_list):
                if self.is_all_latin(el):
                    tagged_tuples.append((el, u"FW"))
                else:
                    tagged_tuples.append((el, postags[i]))
        return tagged_tuples




from num2fawords import words
import re


class Num2String():

    def __init__(self):
        self.number_rules = r'( +| -| )(((\d{1,30})?[./]\d{1,30} )|(\d{1,30} ))'

    def normalize(self, doc_string):
        normalized_string = self.find_number_part(text_line=doc_string)

        return normalized_string

    def find_number_part(self, text_line):
        text_line = " "+text_line+" "

        content_new = re.sub(self.number_rules, lambda x: self.number_converter(x.group()),text_line )
        content_new = re.sub(' +',' ',content_new)
        return content_new.lstrip().strip()

    def number_converter(self, input_data):

        try:
            return   " "+words(input_data, positive='', negative='منهای ', fraction_separator=' ممیز ',
                               ordinal_denominator=False, decimal_separator=' ممیز ') +" "

        except ValueError:
            return input_data



import nltk

class FindChunks():
    def __init__(self):
        self.grammar = r"""
                        VP: {<ADJ_SIM><V_PRS>}
                        VP: {<ADJ_INO><V.*>}
                        VP: {<V_PRS><N_SING><V_SUB>}
                        NP: {<N_SING><ADJ.*><N_SING>}
                        NP: {<N.*><PRO>}
                        VP: {<N_SING><V_.*>}
                        VP: {<V.*>+}
                        NP: {<ADJ.*>?<N.*>+ <ADJ.*>?}
                        DNP: {<DET><NP>}
                        PP: {<ADJ_CMPR><P>}
                        PP: {<ADJ_SIM><P>}
                        PP: {<P><N_SING>}
                        PP: {<P>*}
                        DDNP: {<NP><DNP>}
                        NPP: {<PP><NP>+}
                        """

        self.cp = nltk.RegexpParser(self.grammar)

    def convert_nestedtree2rawstring(self, tree, d=0):
        s = ''
        for item in tree:
            if isinstance(item, tuple):
                s += item[0] + ' '
            elif d >= 1:
                news = self.convert_nestedtree2rawstring(item, d + 1)
                s += news + ' '
            else:
                tag = item._label
                news = '[' + self.convert_nestedtree2rawstring(item, d + 1) + ' ' + tag + ']'
                s += news + ' '
        return s.strip()

    def chunk_sentence(self, pos_taged_tuples):
        return self.cp.parse(pos_taged_tuples)


#
from spacy_loader import nlp

# print(nlp("در سال 5 ه.ق در ساعت 20:30 به گزارش ایسنا سمینار 30-10-1395 شیمی آلی از امروز ۱۱ شهریور ۱۳۹۶ کتابها در دانشگاه علم و صنعت ایران آغاز"
#                   " به کار کرد. این سمینار ﺆ تا ۱۳ شهریور ادامه مییابد خانم‌ها عل9یبیب *- * - 3.57 خواهدبود داشتیدمیرفتید رفتهباشد رفتهاست نرفتهباشد نرفتهبودهام نیاورید گفتن کتابهایت"
#           ))
# print("در سال 5 ه.ق در ساعت 20:30 به گزارش ایسنا سمینار 30-10-1395 شیمی آلی از امروز ۱۱ شهریور ۱۳۹۶ کتابها در دانشگاه علم و صنعت ایران آغاز"
#                   " به کار کرد. این سمینار ﺆ تا ۱۳ شهریور ادامه مییابد خانم‌ها عل9یبیب *- * - 3.57 خواهدبود داشتیدمیرفتید رفتهباشد رفتهاست نرفتهباشد نرفتهبودهام نیاورید گفتن کتابهایت")
# print(nlp.pipe_names)



with nlp.select_pipes(enable=["char_normalizer","affix2norm"]):
    print(nlp("کتابتون رو به من میدهید؟"))
nlp.remove_pipe("spell_checker")
print(nlp.pipeline)

# # nlp.to_disk("./")
# # tokenizer = nlp.tokenizer
# # tokenizer.to_disk("./tokenizer")
# # print(nlp.config.to_str())

# nlp.config.to_disk("./config.json")

#
# nlp.to_disk("./")
# nlp.config.to_disk("./config.json")


from re import sub
import os
from .tokenizer import Tokenizer
from .data_helper import DataHelper
from .token_merger import ClassifierChunkParser
import re
from .stemmer import FindStems
import json


class AffixNorm():

    def __init__(self, config_file=os.path.dirname(os.path.realpath(__file__)) + "/config/affix.json",

                 statistical_space_correction=False,
                 train_file_path="resource/tokenizer/Bijan_khan_chunk.txt",
                 token_merger_path="resource/tokenizer/TokenMerger.pckl",
                 ):
        self.dir_path = os.path.dirname(os.path.realpath(__file__)) + "/"

        self.statistical_space_correction = statistical_space_correction

        with open(config_file) as f:
            self.affix_list = json.load(f)

        self.data_helper = DataHelper()
        self.token_merger = ClassifierChunkParser()

        self.tokenizer = Tokenizer()
        self.stemmer = FindStems(config_file=config_file)

        if self.statistical_space_correction:
            self.token_merger_path = self.dir_path + token_merger_path
            self.train_file_path = train_file_path

            if os.path.isfile(self.token_merger_path):
                self.token_merger_model = self.data_helper.load_var(self.token_merger_path)
            elif os.path.isfile(self.train_file_path):
                self.token_merger_model = self.token_merger.train_merger(self.train_file_path, test_split=0)
                self.data_helper.save_var(self.token_merger_path, self.token_merger_model)

    def space_correction(self, doc_string):

        a00 = r'^(' + "|".join(self.affix_list["space_jointer_1"]) + r')( )'
        b00 = r'\1‌'
        c00 = sub(a00, b00, doc_string)
        a0 = r'( )(' + "|".join(self.affix_list["space_jointer_1"]) + r')( )'
        b0 = r'\1\2‌'
        c0 = sub(a0, b0, c00)
        a1 = r'( )(' + "|".join(self.affix_list["space_jointer_2"]) + r')( )'
        b1 = r'‌\2\3'
        c1 = sub(a1, b1, c0)
        a2 = r'( )' + "|".join(self.affix_list["space_jointer_3"]) + r'( )'
        b2 = r'‌\2‌'
        c2 = sub(a2, b2, c1)
        a3 = r'( )(' + "|".join(self.affix_list["space_jointer_4"]) + r')( )'
        b3 = r'‌\2\3'
        c3 = sub(a3, b3, c2)
        return c3

    def normalize(self, doc_string):

        doc_string = self.space_correction(doc_string)

        normalized_string = self.convert_space(text_line=doc_string)

        return normalized_string

    def convert_space(self, text_line):
        text_line_list = text_line.split()

        text = ' '
        for word in text_line_list:
            result = self.stemmer.convert_to_stem(word)
            text += result + ' '

        text = re.sub(r' +', ' ', text)
        return "".join(text.rstrip().lstrip())



import re


class Tokenizer():
    def __init__(self):
        pass

    def tokenize_words(self, doc_string):
        token_list = doc_string.strip().split()
        token_list = [x.strip("\u200c") for x in token_list if len(x.strip("\u200c")) != 0]
        return token_list

    def tokenize_sentences(self, doc_string):
        #finding the numbers
        pattern = r"[-+]?\d*\.\d+|\d+"
        nums_list = re.findall(pattern, doc_string)
        doc_string = re.sub(pattern, 'floatingpointnumber', doc_string)

        pattern = r'([!\.\?؟]+)[\n]*'
        tmp = re.findall(pattern, doc_string)
        doc_string = re.sub(pattern, self.add_tab, doc_string)

        pattern = r':\n'
        tmp = re.findall(pattern, doc_string)
        doc_string = re.sub(pattern, self.add_tab, doc_string)

        pattern = r';\n'
        tmp = re.findall(pattern, doc_string)
        doc_string = re.sub(pattern, self.add_tab, doc_string)

        pattern = r'؛\n'
        tmp = re.findall(pattern, doc_string)
        doc_string = re.sub(pattern, self.add_tab, doc_string)

        pattern = r'[\n]+'
        doc_string = re.sub(pattern, self.add_tab, doc_string)

        for number in nums_list:
            pattern = 'floatingpointnumber'
            doc_string = re.sub(pattern, number, doc_string, 1)

        doc_string = doc_string.split('\t\t')
        doc_string = [x for x in doc_string if len(x) > 0]
        return doc_string

    def add_tab(self, mystring):
        mystring = mystring.group()  # this method return the string matched by re
        mystring = mystring.strip(' ')  # ommiting the whitespace around the pucntuation
        mystring = mystring.strip('\n') # ommiting the newline around the pucntuation
        mystring = " " + mystring + "\t\t"  # adding a space after and before punctuation
        return mystring


import nltk
from nltk.chunk import conlltags2tree, tree2conlltags
#from sklearn.model_selection import train_test_split
from collections import Iterable
from nltk import ChunkParserI, ClassifierBasedTagger


class ClassifierChunkParser(ChunkParserI):
    def __init__(self):
        self.tagger = None
        pass

    def parse(self, tagged_sent):
        chunks = self.tagger.tag(tagged_sent)
        iob_triplets = [(w, t, c) for ((w, t), c) in chunks]

        # Transform the list of triplets to nltk.Tree format
        return conlltags2tree(iob_triplets)

    def train_merger(self, train_file_path, test_split=0.1):
        print("Loading Data...")
        file = open(train_file_path, "r", encoding='utf-8')
        file_content = file.read()
        file_content = file_content.split("\n\n")

        data_list = []
        for line in file_content:
            line = nltk.chunk.util.conllstr2tree(line, chunk_types=('NP',), root_label='S')
            if (len(line) > 0):
                data_list.append(line)

        # train_sents, test_sents = train_test_split(data_list, test_size=test_split, random_state=91)
        train_sents = data_list
        test_sents = []

        print("Training the model ...")

        # Transform the trees in IOB annotated sentences [(word, pos, chunk), ...]
        chunked_sents = [tree2conlltags(sent) for sent in train_sents]

        # Transform the triplets in pairs, make it compatible with the tagger interface [((word, pos), chunk), ...]
        def triplets2tagged_pairs(iob_sent):
            return [((word, pos), chunk) for word, pos, chunk in iob_sent]

        chunked_sents = [triplets2tagged_pairs(sent) for sent in chunked_sents]

        self.feature_detector = self.features
        self.tagger = ClassifierBasedTagger(
            train=chunked_sents,
            feature_detector=self.features)

        token_merger_model = self.tagger

        if len(test_sents) > 0:
            print("evaluating...")
            print(token_merger_model.evaluate(test_sents))

        return token_merger_model

    def nestedtree_to_list(self, tree, separator_char, d=0):
        s = ''
        for item in tree:
            if isinstance(item, tuple):
                s += item[0] + separator_char
            elif d >= 1:
                news = self.nestedtree_to_list(item, separator_char, d + 1)
                s += news + separator_char
            else:
                news = self.nestedtree_to_list(item, separator_char, d + 1) + '\t'
                s += news + separator_char
        return s.strip(separator_char)

    def merg_tokens(self, token_list, token_merger_model, separator_char):
        # gets a string line as input and returns a list of tokens
        tmp_list = []
        for word in token_list:
            tmp_list.append((word, 'N'))

        self.tagger = token_merger_model

        res = self.parse(tmp_list)
        res = self.nestedtree_to_list(res, separator_char=separator_char)
        res = res.strip('\t').split('\t')
        res = [x.strip(separator_char).strip() for x in res]
        return res

    def features(self, tokens, index, history):
        """
        `tokens`  = a POS-tagged sentence [(w1, t1), ...]
        `index`   = the index of the token we want to extract features for
        `history` = the previous predicted IOB tags
        """

        # Pad the sequence with placeholders
        tokens = [('__START2__', '__START2__'), ('__START1__', '__START1__')] + list(tokens) + [
            ('__END1__', '__END1__'),
            ('__END2__', '__END2__')]
        history = ['__START3__', '__START2__', '__START1__'] + list(history)

        # shift the index with 2, to accommodate the padding
        index += 2

        word, pos = tokens[index]
        prevword, prevpos = tokens[index - 1]
        prevprevword, prevprevpos = tokens[index - 2]
        nextword, nextpos = tokens[index + 1]
        nextnextword, nextnextpos = tokens[index + 2]
        previob = history[index - 1].split("-")[0]
        prevpreviob = history[index - 2].split("-")[0]
        prevprevpreviob = history[index - 3].split("-")[0]

        return {
            'word': word,

            'prev-iob': previob,
            # 'prev-prev-iob' : prevpreviob,
            # 'prev_prev-prev-iob' : prevprevpreviob,

            'next-word': nextword,
            # 'next-next-word': nextnextword,

            'prev-word': prevword,
            # 'prev-prev-word': prevprevword,
        }